function propagatelabel(srcI3, initLabel, initLabelRGB, numOfLabels, motModel, apModel, enMinOpt, save2hdd, varargin)
% propagatelabel(srcI3, initLabel, initLabelRGB, numOfLabels, motModel, apModel, enMinOpt, save2hdd, varargin)
%   takes the space-time volume (srcI3) and the labels (initLabel, initLabelRGB)
%   and propagates the labels throughout the whole video. Propagation
%   results are stored directly to the hard drive.
%
% srcI3:    the space-time volume of the to-be-propagated section of the video
%           (srcI3 should be RGB with type UINT8!!!)
%           (RGB will be converted to Lab before learning the appearance model)
% initLabel:    pixel-wise labels (scaler values) for the first frame 
% initLabelRGB: pixel-wise labels (RGB values) for the first frame
% motModel: 0. No motion model used at all
%           1. M-1.1 Forward flow only 
%           2. M-1.2 Reverse flow only
%           3. M-2.1.0 Label prob. map from forward flow, hard valued
%           4. M-2.1.1 Label prob. map from forward flow, soft valued
%           5. M-2.1.2 Label prob. map from forward and reverse flow (soft)
%
% apModel:  0. No appearance model used at all
%           1. single pixel grayscale
%           2. single pixel Lab colorspace
%           3. patch grayscale
%           4. patch Lab colorspace
%           5. superpixel Lab colorspace
%           8. learning multiple oversegmented appearance models for each
%              semantic label, then track them over time and update them.
% 
% enMinOpt: 0: don't touch the outputs at all 
%           1. morphological smoothing, max likelihood.
%           2. MRF, fixed weights between energy terms
%           3. M-3.1 MRF, automatically adjusted weights for each pixel,
%              ver. 1
%           31. MRF, automatically adjust weights for each pixel, ver2.
%
% save2hdd: 0. don't save at all
%           1. save propagated labels
%           2. save the alpha blended labels
%           3. save both the propahated labels and the alpha blended labels
%
% varargin: optional arguement for passing pre-calculated optical flows.
%           If srcI3 contains frame 1-75, then 
%           varargin{1} are the forward flows originated from frames 1-74.
%           varargin{2} are the reverse flows originated from frames 2-75.
%           if both forward and reverse flow are used.
%           Otherwise, only varargin{1} would be forward or reverse flow,
%           depending on the motModel used.
%
%
%
% Author: Albert Y. C. Chen, SUNY at Buffalo, USA (aychen@buffalo.edu)
% $Date: 2011-02-15$
% $Revision: 0 $
%
% Copyright 2011, SUNY at Buffalo, Buffalo, NY USA
% 
%                         All Rights Reserved
% 
% Permission to use, copy, modify, and distribute this software and its
% documentation for any purpose other than its incorporation into a
% commercial product is hereby granted without fee, provided that the
% above copyright notice appear in all copies and that both that
% copyright notice and this permission notice appear in supporting
% documentation, and that the name of the author and SUNY Buffalo not be 
% used in advertising or publicity pertaining to distribution of the 
% software without specific, written prior permission.
% 
% THE AUTHOR AND SUNY BUFFALO DISCLAIM ALL WARRANTIES WITH REGARD TO THIS 
% SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
% FOR ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR OR SUNY BUFFALO
% BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY 
% DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER 
% IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING 
% OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.   

displayImages=1;

% parse the input arguments
if motModel~=0 & motModel~=1 & motModel~=2 & motModel~=3 & motModel~=4 & motModel~=5
    error('Mode not supported! type help propagatelabelv6 for instructions.')
end
if apModel~=0 & apModel~=1 & apModel~=2 & apModel~=3 & apModel~=4 & apModel~=5 & apModel~=8
    error('Mode not supported! type help propagatelabelv6 for instructions.')
end
if enMinOpt~=0 & enMinOpt~=1 & enMinOpt~=2 & enMinOpt~=3 & enMinOpt~=4 & enMinOpt~=31
    error('Mode not supported! type help propagatelabelv6 for instructions.')
end
if motModel==0 & apModel==0
    error('At least one of the two, motModel and apModel, should be non zero.')
end
optargin = size(varargin,2);
if optargin==2
    listV = varargin{1};
    listVr = varargin{2};
elseif optargin==1
    if (motModel==1 | motModel==3 | motModel==4)
        listV = varargin{1};
    elseif (motModel==2)
        listVr = varargin{1};
    elseif (motModel==5)
        error('While using motModel==5 (joint forward and reverse flow), please give 0 or 2 optional arguments,')
    else
        error('This should never happen..., since the input parameters are safe-guarded already.')
    end
end


disp(sprintf('===== Label Propagation ver. 0.8. =====\n (motModel = %d, apModel = %d, enMinOpt = %d)\n', motModel, apModel, enMinOpt));

% set the paths to access all the sub-directories where this file is
% currently located. Required for the 
addpath(genpath('.'));

% constant parameters
alpha = 0.5;
numOfBins = 32;
patchRadius = 4;


% take care of the windows for displaying stuff
if displayImages==1
    close all;
    figure; % input image
    colormap gray;
    figure; % RGB labels 
    colormap gray;
    figure; % optical flows (if any)
    colormap gray;
end

% get the input image paramters and dummy proof them
numOfFrames = size(srcI3,2);
[rows columns srcImgDim] = size(srcI3{1});
if srcImgDim~=3
    error('Each cell of srcI3{i} should be a m*n*3 RGB image!');
end

% convert srcI3 into the appropriate colorspace as I3
I3 = cell(1,numOfFrames);
if apModel==2 | apModel==4 | apModel==5 | apModel==8
    cform = makecform('srgb2lab');
    for i=1:numOfFrames
        I3{i} = applycform(srcI3{i},cform);
    end
elseif apModel==1 | apModel==3
    I3 = cell(1,numOfFrames);
    for i=1:numOfFrames    
    	I3{i} = rgb2gray(srcI3{i});
    end
end
dim = size(I3{1}, 3);

% create the data structure used for recording the appearance model
if apModel~=0 & apModel~=8
    apModelNorm = zeros(numOfBins, numOfLabels, dim);
    apModelAccum = zeros(numOfBins, numOfLabels, dim);
end

% create the mapping between label and labelRGB
% the mapping is stored as a 4xN array, where each column is a mapping
% while row 1 is the scalar value, row 2-4 are the r,g,b values
% respectively
uniqueLabels = unique(initLabel);
colorMapping = cell(1,3);
for i=1:3
    colorMapping{i} = zeros(numOfLabels,1);
end
for i=1:size(uniqueLabels,1)
    mask = repmat(initLabel==uniqueLabels(i), [1,1,3]);
    correspondingRgbLabels = unique(reshape(initLabelRGB(mask),[],3), 'rows');
    if size(correspondingRgbLabels,1)~=1
        error('the pixel lables in "initLabel" and "initLabelRGB" should be one-on-one!');
    end
    colorMapping{1}(uniqueLabels(i)+1) = correspondingRgbLabels(1);
    colorMapping{2}(uniqueLabels(i)+1) = correspondingRgbLabels(2);
    colorMapping{3}(uniqueLabels(i)+1) = correspondingRgbLabels(3);
end

% create other data structures
if motModel~=2 & motModel~=0
    numOfZeroProposals = zeros(numOfFrames, 1);
    numOfMultiProposals = zeros(numOfFrames, 1);
end

nextFrameLabelRGB = initLabelRGB;
nextFrameLabel = initLabel;

for i=1:numOfFrames-1    
    disp(sprintf('Propagating labels from frame I3{%d} to I3{%d}.\n', i, i+1));
    %======================================================================
    % STEP 1: show the image, labels, and the optical flow (calc. if ness.)
    %======================================================================
    
    % update stuff...
    currentFrameLabel = nextFrameLabel;
    currentFrameLabelRGB = nextFrameLabelRGB;
    nextFrameLabel = zeros(rows, columns);
    nextFrameLabelRGB = zeros(rows, columns, 3);
    
    % show the original input image in fugure 1 
    if displayImages==1
        figure(1);
        imshow(srcI3{i});
    end

    % show the image (grayscal version) alpha blended with the labels 
    % (user-provided or calculated in the previous round) in figure 2
    
    if srcImgDim==3
        img = rgb2gray(srcI3{i});
    elseif srcImgDim==1
        img = srcI3{i};
    end
    blended(:,:,1) = currentFrameLabelRGB(:,:,1).*(1-alpha) + img.*alpha;
    blended(:,:,2) = currentFrameLabelRGB(:,:,2).*(1-alpha) + img.*alpha;
    blended(:,:,3) = currentFrameLabelRGB(:,:,3).*(1-alpha) + img.*alpha;
    clear img;
    if displayImages==1
        figure(2);
        imshow(blended);
    end
    if save2hdd==1 | save2hdd==3
        imwrite(blended, sprintf('propagated_blended_%06d.png', i-1), 'png');
    end
    if save2hdd==2 | save2hdd==3
        imwrite(currentFrameLabelRGB, sprintf('propagated_%06d.png', i-1), 'png');
    end
    % show (and calculate if necessary) the optical flows in figure 3
    % V in oflow are directed from frame i to frame i+1.
    if motModel~=0
        if motModel==1 | motModel==3 | motModel==4 | motModel==5
            if optargin==0
                disp(sprintf('Calculating the forward optical flows between frame I3{%d} and I3{%d}.\n', i, i+1));
                V = estimate_flow_ba(srcI3{i}, srcI3{i+1});
            else
                disp(sprintf('Using the precalculated forward optical flows between frame I3{%d} and I3{%d}.\n', i, i+1));
                V = listV{i};
            end
        end
        if motModel==2 | motModel==5
            if optargin==0
                disp(sprintf('Calculating the reverse optical flows between frame I3{%d} and I3{%d}.\n', i, i+1));
                Vr = estimate_flow_ba(srcI3{i+1}, srcI3{i});
            else
                disp(sprintf('Using the precalculated reverse optical flows between frame I3{%d} and I3{%d}.\n', i, i+1));
                Vr = listVr{i+1};
            end
        end

        if displayImages==1
            figure(3);
            imshow(srcI3{i});
            hold on;
            s=10;
            [x y] = meshgrid(1:s:columns,1:s:rows);
            if motModel~=2
                quiver(x,y,V(1:s:rows,1:s:columns,1),V(1:s:rows,1:s:columns,2),1);
            else
                quiver(x,y,Vr(1:s:rows,1:s:columns,1),Vr(1:s:rows,1:s:columns,2),1);
            end
            axis ij
        end
    end    

    %For the SACK OF SUPERPIXELS (SSP) method
    if apModel==8 & i==1
        bin0 = 0;
        bin1 = 0;
        %%%% Obtain the SSP for the current labeled frame
        sspNum = 0;
        % Go through each of the semantic segments and break it into 
        % a sack of superpixels
        ssp = uint16(zeros(rows, columns));
        for l = uniqueLabels' %"uniqueLabels" need to be a row vector
           % get the mask for the lth semantic segment
           mask = uint8(currentFrameLabel==l);
           mask3d(:,:,1) = mask; mask3d(:,:,2) = mask; mask3d(:,:,3) = mask; 
           % dilute mask3d, s.t. we can get the surrounding region of
           % the current semantic segment, so that small over-segmented 
           % superpixels don't appear near the boundary.
           SE = strel('disk',3);
           dilatedMask3d = imdilate(mask3d, SE);
           dilatedCurrSeg = I3{i}.*dilatedMask3d;
           [currSegSSPColor, currSegSSP] = segment(dilatedCurrSeg, 1, 150, 300);
           % modify the ssp number assigned to the current segment's ssp
           % according to the existing # of ssp's.
           currSegSSP = (imclose(uint16(currSegSSP), SE)+sspNum).*uint16(mask); %note that the indices are not in sequential order
           sspNum = sspNum + max(unique(currSegSSP)) + 1; 
           % Add these superpixels to the final sack of superpixel matrix 
           ssp = ssp + currSegSSP;
           
           
        end
        %modify the 
        
        % ssp is stored internally as uint8, yet outputted as uint8 for the
        % ease of viewing
        imwrite(uint8(ssp), sprintf('ssp_frame%06d.png', i), 'png');
        
        
        %%%% Learn the appearance model for each of the SSP
        sspIdx = unique(ssp);
        numOfUniqueSSP = size(sspIdx, 1);
        apModelNorm = zeros(numOfBins, dim, numOfUniqueSSP);
        apModelAccum = zeros(numOfBins, dim, numOfUniqueSSP);
        sspNum2Label = zeros(numOfUniqueSSP, 1);
        
        for s = 1:numOfUniqueSSP
            for d = 1:dim
                % learn the appearance model
                colorChannel = I3{i}(:,:,d);
                [counts bin] = imhist(colorChannel(ssp==sspIdx(s)), numOfBins);
                apModelAccum(:,d,s) = apModelAccum(:,d,s) + counts; 
                apModelNorm(:,d,s) = apModelAccum(:,d,s)/sum(apModelAccum(:,d,s));
            end
            % create the correspoding mapping between ssp# and label#
            sspIdx2Label(s) =  unique(currentFrameLabel(ssp==sspIdx(s)));
        end
    end
    
    
    %======================================================================
    % Propogate the labels from frame i to i+1 using motion, appearance,
    % and spatial smoothness information
    %
    % The user has the option to output the maximum likelihood results here
    % directly or output the unary term(s) to external energy minimizers.
    %======================================================================
    
    %%%% U^A Term (unary appearance) %%%%
    % learn or update the appearance model from the labeled images.
    % Don't learn for the 0 class (void/catch all class)
    % (learn a histogram of the intensity values as a start)
    % class 0-23 are stored in apModel*(j), j=1:24
    if apModel~=0 & apModel~=8
        uniqueLabels = unique(currentFrameLabel);
        % uniqueLabels(uniqueLabels==0) = []; % optional, get rid of the null label
        for l=uniqueLabels' % Assuming that labels start from 0---therefore, store label l at the (l+1)th position
            for d=1:dim
                colorChannel = I3{i}(:,:,d);
                [counts bin] = imhist(colorChannel(currentFrameLabel==l), numOfBins);
                apModelAccum(:,l+1,d) = apModelAccum(:,l+1,d) + counts; % update the model at each frame
                apModelNorm(:,l+1,d) = apModelAccum(:,l+1,d)/sum(apModelAccum(:,l+1,d));
            end
        end
    end
    
    
    %%%% U^M Term (unary motion) %%%%
    if motModel~=0
        % nextFrameLabelProposal is the core structures of this program.
        % all the possible labels (temporal neighbors) proposed are stored here
        % nextFrameLabelProposal{y, x}{:,n} = info of nth proposal for pixel (x,y)
        % where {1,n} stores the (row,column) coordinate of the proposer
        %       {2,n} records the proposed label
        %       {3,n} stores the confidence value
        clear nextFrameLabelProposal;
        nextFrameLabelProposal = cell(rows, columns);

        % Forward Optical Flow: Go through all x,y locations in frame t and
        % record the x,y of where its optical flow terminates in frame t+1.
        if motModel==1 | motModel==3 | motModel==4 | motModel==5
            [x y] = meshgrid(1:columns,1:rows);
            x = round(x+V(:,:,1));
            y = round(y+V(:,:,2));
            x(x>columns)=columns;
            y(y>rows)=rows;
            x(x<1)=1;
            y(y<1)=1;
        end
        % Reverse Optical Flow: 
        if motModel==2 | motModel==5
            [xr yr] = meshgrid(1:columns,1:rows);
            xr = round(xr+Vr(:,:,1));
            yr = round(yr+Vr(:,:,2));
            xr(xr>columns) = columns;
            yr(yr>rows) = rows;
            xr(xr<1) = 1;
            yr(yr<1) = 1;
        end

        % Pre-process the foward flows. Store them in "nextFrameLabelProposal"
        % For every pixel (j,k) in frame i, find its outgoing optical flow's 
        % termination point (y(j,k), x(j,k)) in frame i+1. For every pixel
        % (y(j,k), x(j,k)), label proposals are stored column-wise as:
        % 1st row: location; 2nd row: proposed label; 3rd row: confidence value
        if motModel==1 | motModel==3 | motModel==4 | motModel==5
            % go through all x, y locations in frame i
            for k=1:columns
                for j=1:rows
                    % add the inverse flow's proposal to the list first.
                    numOfProposals = size(nextFrameLabelProposal{y(j,k), x(j,k)}, 2);
                    nextFrameLabelProposal{y(j,k), x(j,k)}{1, numOfProposals+1} = [j, k];
                    nextFrameLabelProposal{y(j,k), x(j,k)}{2, numOfProposals+1} = currentFrameLabel(j, k);
                    if motModel==3
                        nextFrameLabelProposal{y(j,k), x(j,k)}{3, numOfProposals+1} = 1; 
                    elseif motModel==4 | motModel==5
                        % Get the subimage surrounding the proposer and its subject to
                        % calculate the energy of the optical flow
                        % proposer: patchBounds(1,:), subject: patchBounds(2,:)
                        patchBounds(2,:) = [y(j,k)-patchRadius, y(j,k)+patchRadius, x(j,k)-patchRadius, x(j,k)+patchRadius];
                        patchBounds(1,:) = [j-patchRadius, j+patchRadius, k-patchRadius, k+patchRadius];
                        patchBounds(patchBounds<1)=1; % make sure the subimages are within the boundaries of the image
                        patchBounds([logical(zeros(2,1)), patchBounds(:,2)>rows, logical(zeros(2,2))]) = rows;
                        patchBounds([logical(zeros(2,3)), patchBounds(:,4)>columns]) = columns;
                        % Get the normalized histogram of the proposer and its subject.
                        % Measure the similarity of the two and record it as the
                        % confidence value of the optical flow.
                        confidence = 1;
                        for d=1:srcImgDim
                            [proposerHist, bin] = imhist(srcI3{i}(patchBounds(1,1):patchBounds(1,2), patchBounds(1,3):patchBounds(1,4), d), numOfBins);
                            [subjectHist, bin] = imhist(srcI3{i+1}(patchBounds(2,1):patchBounds(2,2), patchBounds(2,3):patchBounds(2,4), d), numOfBins);
                            proposerHist = proposerHist/sum(proposerHist);
                            subjectHist = subjectHist/sum(subjectHist);
                            confidence = confidence * sum(min(proposerHist, subjectHist)); % a simple intersection measure. 1: total match. 0: total dismatch
                        end
                        nextFrameLabelProposal{y(j,k), x(j,k)}{3, numOfProposals+1} = confidence; 
                    end
                end
            end
        end
    end
    
    %%%%
    % Compute the likelihood maps for the motion/appearance terms
    %%%%
    if motModel==1 | motModel==3 | motModel==4 | motModel==5
        labelProposalMap_zero = logical(zeros(rows, columns));
        labelProposalMap_multi = logical(zeros(rows, columns));
        labelProposalMap_single = logical(zeros(rows, columns));
    end
    if apModel~=0
        probMapAppearance = zeros(rows, columns, numOfLabels);
    end
    if motModel~=0
        probMapMotion = zeros(rows, columns, numOfLabels);
    end

    %===============================================================
    % Generate the label likelihood maps from motion and appearance
    %===============================================================
    
    % pre-process the superpixel and appearance histograms if necessary
    if apModel==5 | apModel==8
        % Call the "Efficient Graph-Based Image Segmentation" mex file "segment"
        % in the /segment folder.
        [spColor, spScalar] = segment(I3{i+1},1,20,50);
        maxSegmentIndex = max(spScalar(:));
        % precalculate and store the maxSegmentIndex number of appearance
        % histograms, to avoid having to do this e.g. 320*240 times
        apHistSegment = zeros(numOfBins, dim, maxSegmentIndex+1);
        for spidx=0:maxSegmentIndex 
            for d=1:dim
                colorChannel = I3{i+1}(:,:,d);
                [counts bin] = imhist(colorChannel(spScalar==spidx), numOfBins);
                apHistSegment(:,d,spidx+1) = counts/sum(counts);
            end
        end
    end
    
    % the weight maps are for dynamically adjuesting the weight between
    % the different unary terms, depending on the reliability of the 
    % optical flow information.
    if enMinOpt==3
        weightMapAppearance = zeros(rows, columns);
        weightMapMotion = zeros(rows, columns);
    end
    % go through all x,y locations in frame i+1
    for k=1:columns
        for j=1:rows
            % add the inverse flow to the proposal list, last.
            if motModel==5
                numOfProposals = size(nextFrameLabelProposal{j, k}, 2);
                nextFrameLabelProposal{j,k}{1, numOfProposals+1} = [yr(j,k), xr(j,k)];
                nextFrameLabelProposal{j,k}{2, numOfProposals+1} = currentFrameLabel(yr(j,k), xr(j,k));
                % Calculate the optical flow confidence value for the
                % inverse flow, same as those used for forward flows
                patchBounds(2,:) = [yr(j,k)-patchRadius, yr(j,k)+patchRadius, xr(j,k)-patchRadius, xr(j,k)+patchRadius];
                patchBounds(1,:) = [j-patchRadius, j+patchRadius, k-patchRadius, k+patchRadius];
                patchBounds(patchBounds<1)=1; % make sure the subimages are within the boundaries of the image
                patchBounds([logical(zeros(2,1)), patchBounds(:,2)>rows, logical(zeros(2,2))]) = rows;
                patchBounds([logical(zeros(2,3)), patchBounds(:,4)>columns]) = columns;
                confidence = 1; %use the inverse of the diff. of two hist. as conf. value
                for d=1:srcImgDim
                    [proposerHist, bin] = imhist(srcI3{i+1}(patchBounds(1,1):patchBounds(1,2), patchBounds(1,3):patchBounds(1,4), d), numOfBins);
                    [subjectHist, bin] = imhist(srcI3{i}(patchBounds(2,1):patchBounds(2,2), patchBounds(2,3):patchBounds(2,4), d), numOfBins);
                    proposerHist = proposerHist/sum(proposerHist);
                    subjectHist = subjectHist/sum(subjectHist);
                    confidence = confidence * sum(min(proposerHist, subjectHist)); % a simple intersection measure. 1: total match. 0: total dismatch
                end
                nextFrameLabelProposal{j,k}{3, numOfProposals+1} = confidence; 
            end
            %---------------------------------------
            % compute the appearance likelihood map
            %---------------------------------------
            if apModel~=0
                apProb = zeros(numOfLabels, 1);
                if apModel==1 | apModel==2 %single pixel value-based
                    for l=1:numOfLabels
                        apProb(l) = 1;
                        for d=1:dim
                            bin = floor(double(I3{i+1}(j,k,d))/numOfBins)+1;
                            apProb(l) = apProb(l)*apModelNorm(bin, l, d);
                        end
                    end
                    if sum(apProb)==0
                        probMapAppearance(j,k,:) = ones(1,numOfLabels)*(1/numOfLabels);
                    else
                        probMapAppearance(j,k,:) = apProb/sum(apProb);
                    end
                elseif apModel==3 | apModel==4 % patch histogram-based
                    % Get the subimage surrounding the proposer and its subject to
                    % calculate the energy of the optical flow
                    % proposer: patchBounds(1,:), subject: patchBounds(2,:)
                    patchBounds2 = [j-patchRadius, j+patchRadius, k-patchRadius, k+patchRadius];
                    patchBounds2(patchBounds2<1)=1; % make sure the subimages are within the boundaries of the image
                    patchBounds2([logical(zeros(1,1)), patchBounds2(2)>rows, logical(zeros(1,2))]) = rows;
                    patchBounds2([logical(zeros(1,3)), patchBounds2(4)>columns]) = columns;                    
                    for d=1:dim
                        [subImageHist(:,d), bin] = imhist(I3{i+1}(patchBounds2(1):patchBounds2(2), patchBounds2(3):patchBounds2(4), d), numOfBins);
                        subImageHist(:,d) = subImageHist(:,d)/sum(subImageHist(:,d));
                    end
                    for l=1:numOfLabels
                        apProb(l) = 1;
                        for d=1:dim
                            prob = sum(min(subImageHist(:,d), apModelNorm(:,l,d)));
                            apProb(l) = apProb(l)*prob;
                        end
                    end
                    if sum(apProb)==0
                        probMapAppearance(j,k,:) = ones(1,numOfLabels)*(1/numOfLabels);
                    else
                        probMapAppearance(j,k,:) = apProb/sum(apProb);
                    end
                elseif apModel==5 %superpixel-based
                    spidx = spScalar(j,k); % get the superpixel's index value
                    %get the pre-computed color histogram for each
                    %superpixel
                    for l=1:numOfLabels
                        apProb(l) = 1;
                        for d=1:dim
                            prob = sum(min(apHistSegment(:,d,spidx+1), apModelNorm(:,l,d)));
                            apProb(l) = apProb(l)*prob;
                        end
                    end     
                    % normalize it!
                    if sum(apProb)==0
                        probMapAppearance(j,k,:) = ones(1,numOfLabels)*(1/numOfLabels);
                    else
                        probMapAppearance(j,k,:) = apProb/sum(apProb);
                    end
                elseif apModel==8 %super-pixel with SSP appearance
                    spidx = spScalar(j,k); % get the superpixel's index value
                    % compute the per SSP probability 
                    sspProb = zeros(numOfUniqueSSP, 1);
                    for s = 1:numOfUniqueSSP
                        prob = 1;
                        for d = 1:dim
                            prob = prob * sum(min(apHistSegment(:,d,spidx+1), apModelNorm(:,d,s)));
                        end
                        sspProb(s) = prob;
                    end     
                    % convert the per SSP probability into per class
                    % probaility 
                    for l= 1:numOfLabels
                        % each semantic segment can be composed of multiple
                        % ssp. Each ssp has its own appearance probability.
                        % We take the max probability one.
                        sspMaxProb = max(sspProb(sspIdx2Label==(l-1)));
                        if isempty(sspMaxProb)
                            apProb(l) = 0;
                        else
                            apProb(l) = sspMaxProb;
                        end
                    end
                    % normalize it!
                    if sum(apProb)==0
                        probMapAppearance(j,k,:) = ones(1,numOfLabels)*(1/numOfLabels);
                    else
                        probMapAppearance(j,k,:) = apProb/sum(apProb);
                    end
                end
            end        
            
            %-----------------------------------
            % compute the motion likelihood map
            %-----------------------------------
            if motModel==1 %forward flow only
                numOfProposals = size(nextFrameLabelProposal{j, k}, 2);
                if numOfProposals<=0
                    numOfZeroProposals(i) = numOfZeroProposals(i)+1;
                    labelProposalMap_zero(j, k) = 1;
                    nextFrameLabel(j,k) = 0;
                elseif numOfProposals>=2
                    nextFrameLabel(j,k) = nextFrameLabelProposal{j,k}{2,1};
                    numOfMultiProposals(i) = numOfMultiProposals(i)+1;
                    labelProposalMap_multi(j, k) = 1;
                else %numOfProposals==1
                    nextFrameLabel(j,k) = nextFrameLabelProposal{j,k}{2,1};
                end 
            elseif motModel==2 %inverse flow only
                nextFrameLabel(j, k) = currentFrameLabel(yr(j, k), xr(j, k));
            elseif motModel==3 | motModel==4 | motModel==5
                numOfProposals = size(nextFrameLabelProposal{j, k}, 2);
                % don't count the "just added" inverse flows when motModel==5
                if motModel==5 
                    numOfProposals = numOfProposals-1;
                end
                %%%% no incoming forward flows
                if numOfProposals<=0 % could never be true for motModel==5
                    numOfZeroProposals(i) = numOfZeroProposals(i)+1;
                    labelProposalMap_zero(j, k) = 1;
                    probMapMotion(j,k,:) = ones(1, numOfLabels)*(1/numOfLabels);
                    if enMinOpt==3 | enMinOpt==31
                        weightMapMotion(j,k) = exp(-3);
                        weightMapAppearance(j,k) = 1 - exp(-3);
                    end
                %%%% multi incoming forward flow
                elseif numOfProposals>=2 
                    numOfMultiProposals(i) = numOfMultiProposals(i)+1;
                    labelProposalMap_multi(j, k) = 1;

                    % Sum of confidence value of all incoming/outgoing
                    % optical flows
                    motionConfidenceSum = 0;
                    if motModel==5 & enMinOpt==31
                        maxConfIndex = 0;
                        maxConfValue = 0;
                    end
                    for l=1:numOfProposals
                        motionConfidenceSum = motionConfidenceSum + nextFrameLabelProposal{j, k}{3, l};
                        % finding the appropriate temporal anchor point
                        if motModel==5 & enMinOpt==31
                            if nextFrameLabelProposal{j,k}{3,l} > maxConfValue
                                maxConfValue = nextFrameLabelProposal{j,k}{3,l};
                                maxConfIndex = l;
                            end
                        end
                    end
                    if motModel==5 & (enMinOpt==3 | enMinOpt==31)% the inverse flow
                        motionConfidenceSum = motionConfidenceSum + nextFrameLabelProposal{j, k}{3, numOfProposals+1};
                    end
                    
                    % Generate the motion probability maps
                    delta = 0;
                    for l=1:numOfProposals
                        label = nextFrameLabelProposal{j, k}{2, l};
                        %disp(sprintf('[m] (row, column) = (%d, %d); numOfProposals: %d, label: %d, l: %d\n', j, k, numOfProposals, label, l));
                        probMapMotion(j,k,label+1) = probMapMotion(j,k,label+1) + nextFrameLabelProposal{j, k}{3, l}/motionConfidenceSum;
                        
                        % compute weightMapAppearance and weightMapMotion
                        if motModel==5 & enMinOpt==3 %always use the inverse flow as the anchor
                            yDistance = abs(yr(j,k) - nextFrameLabelProposal{j,k}{1,l}(1));
                            xDistance = abs(xr(j,k) - nextFrameLabelProposal{j,k}{1,l}(2));
                            delta = delta + (yDistance+xDistance)/2;
                        elseif motModel==5 & enMinOpt==31 % use the pre-calc. anchor
                            if maxConfIndex~=l
                                yDistance = abs(nextFrameLabelProposal{j,k}{1,l}(1) - nextFrameLabelProposal{j,k}{1,maxConfIndex}(1));
                                xDistance = abs(nextFrameLabelProposal{j,k}{1,l}(2) - nextFrameLabelProposal{j,k}{1,maxConfIndex}(2));
                                delta = delta + (yDistance+xDistance)/2;
                            end
                        end
                    end
                    
                    if motModel==5 & enMinOpt==3 
                        %add the inverse flow's effect into the prob. maps
                        label = nextFrameLabelProposal{j, k}{2, numOfProposals+1};
                        probMapMotion(j,k,label+1) = probMapMotion(j,k,label+1) + nextFrameLabelProposal{j, k}{3, numOfProposals+1}/motionConfidenceSum;
                        
                        % update weightMapAppearance and weightMapMotion 
                        delta = delta/(numOfProposals+1);
                        weightMapMotion(j,k) = exp(-delta);
                        weightMapAppearance(j,k) = 1 - weightMapMotion(j,k);
                    end
                        
                %%%% exactly one incoming forward flow
                else 
                    %disp(sprintf('[1] (row, column) = (%d, %d); numOfProposals: %d, label: %d\n', j, k, numOfProposals, label));
                    labelProposalMap_single(j,k) = 1;
                    label = nextFrameLabelProposal{j, k}{2, 1};
                    probMapMotion(j,k,label+1) = nextFrameLabelProposal{j,k}{3,1};
                    
                    if motModel==5 & (enMinOpt==3 | enMinOpt==31) %which point is the anchor is not important here
                        yDistance = abs(yr(j,k) - nextFrameLabelProposal{j,k}{1,1}(1));
                        xDistance = abs(xr(j,k) - nextFrameLabelProposal{j,k}{1,1}(2));
                        delta = (yDistance+xDistance)/2;
                        weightMapMotion(j,k) = exp(-delta);
                        weightMapAppearance(j,k) = 1 - weightMapMotion(j,k);
                    end
                end
            end
            
            
        end
    end
    
    %================================================================
    % record the probability maps for the next frame by joining the
    % multiple likelihood maps together. (Determined by "enMinOpt").
    %================================================================
    for k=1:numOfLabels
        if apModel~=0
           imwrite(probMapAppearance(:,:,k), sprintf('probMapAppearance_%06d-%d.png', i, k-1), 'png'); 
        end
        if motModel~=0 & motModel~=1 & motModel~=2
            imwrite(probMapMotion(:,:,k), sprintf('probMapMotion_%06d-%d.png', i, k-1), 'png');
        end
    end
    if motModel~=0 & motModel~=2
        imwrite(double(labelProposalMap_zero), sprintf('map_zeroproposal_%06d.png', i), 'png');
        imwrite(double(labelProposalMap_multi), sprintf('map_multiproposal_%06d.png', i), 'png');
    end
    if motModel==5 & enMinOpt==3
        imwrite(weightMapAppearance, sprintf('weightMapAppearance_%06d.png', i), 'png');
        imwrite(weightMapMotion, sprintf('weightMapMotion_%06d.png', i), 'png');
    end
    
    % decide the labels for the next frame
    if apModel==0 & ( motModel==3 | motModel==4 | motModel==5)
        probMap = probMapMotion;
    elseif apModel~=0 & motModel==0
        probMap = probMapAppearance;
    elseif apModel~=0 & ( motModel==3 | motModel==4 | motModel==5)
        if enMinOpt==2 %using global weights
            probMap = (probMapAppearance+probMapMotion)/2;
        elseif enMinOpt==3
            for k=1:numOfLabels
                probMap(:,:,k) = probMapAppearance(:,:,k).*weightMapAppearance + probMapMotion(:,:,k).*weightMapMotion;
            end
        end
    end
    
    if enMinOpt==0 % just take the max. probability results
        if apModel~=0 | (motModel~=1 & motModel~=2)
            [maxProbVal, maxProbIndex] = max(probMap,[],3);
            nextFrameLabel = maxProbIndex-1;
        end
    elseif enMinOpt==1 % morphological smoothing
        nextFrameLabel = uint8(imopen(nextFrameLabel, strel('disk', 1)));
        nextFrameLabel = uint8(imclose(nextFrameLabel, strel('disk', 1))); 
    elseif (enMinOpt==2 | enMinOpt==3) & (motModel==3 | motModel==4 | motModel==5 | apModel~=0)
        nextFrameLabel = calclabel(probMap, uint8(1), 0.1); %call external MRF library
    end

    % color the labels for the next frame
    nextFrameLabelRGB(:,:,1) = colorMapping{1}(nextFrameLabel+1);
    nextFrameLabelRGB(:,:,2) = colorMapping{2}(nextFrameLabel+1);
    nextFrameLabelRGB(:,:,3) = colorMapping{3}(nextFrameLabel+1);
    nextFrameLabelRGB = uint8(nextFrameLabelRGB);
end

